@@ -1,5 +1,5 @@
 ;; cldomain is a Common Lisp domain for the Sphinx documentation tool.
-;; Copyright (C) 2011-2012 Russell Sim <russell.sim@gmail.com>
+;; Copyright (C) 2011-2014 Russell Sim <russell.sim@gmail.com>
 
 ;; This program is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -104,6 +104,11 @@ is a string then just return the string."
 (defun encode-xref (symbol)
   (encode-symbol symbol ":cl:symbol:`~~~a`"))
 
+(defun encode-literal (symbol)
+  (if (equal (symbol-package symbol) (find-package 'keyword))
+      (format nil "``~s``" symbol)
+      (format nil "``~a``" symbol)))
+
 (defun encode-specializer (atom)
   "encode a single specializer lambda list"
   (cond ((eq (type-of atom) 'eql-specializer)
@@ -118,26 +123,54 @@ is a string then just return the string."
   (let ((methods (closer-mop:generic-function-methods generic-function)))
     (with-object ()
       (dolist (method methods)
-        (let ((specializer (closer-mop:method-specializers method)))
+        (let ((specializer (closer-mop:method-specializers method))
+              (lambda-list (closer-mop:method-lambda-list method)))
           (encode-object-member
            (mapcar #'encode-specializer specializer)
            (scope-symbols-in-text
-            (or (documentation method t) ""))))))))
+            (or (documentation method t) "")
+            (simplify-arglist lambda-list))))))))
+
+(defun intern-keyword (keyword)
+  (let ((keyword (if (eql (char keyword 0) #\:)
+                     (subseq keyword 1) keyword)))
+    (intern (string-upcase keyword) 'keyword)))
 
 (defun find-best-symbol (symbols &optional ignore-symbols)
   "try and find the best symbol, return the most appropriate and the
 remaining string.  symbols is a list of strings that contains
 possible symbol names."
   (dolist (symbol-string symbols)
-    (let ((symbol (find-symbol* symbol-string)))
-      (when (and (not (null symbol)) (symbolp symbol))
-        (when (member symbol ignore-symbols)
-          (return))
-        (return-from find-best-symbol
-          (values symbol (if (equal symbol-string (car symbols))
-                           ""  ; first symbol, so no remainder
-                           (subseq (car symbols) (length symbol-string))))))))
-  (values nil (car symbols)))
+    (let ((symbol (find-symbol* symbol-string))
+          (rest-string
+            (if (equal symbol-string (car symbols))
+                ""  ; first symbol, so no remainder
+                (subseq (car symbols) (length symbol-string)))))
+      (cond
+        ;; Exception for keywords with full stops at the end.  It's
+        ;; probably punctuation.
+        ((and (eql (char symbol-string 0) #\:)
+              (member (char symbol-string (1- (length symbol-string)))
+                      (list #\. #\,)))
+         (return-from find-best-symbol
+           (values nil
+                   (intern-keyword (subseq symbol-string
+                                         1
+                                         (1- (length symbol-string))))
+                   (subseq symbol-string (1- (length symbol-string))))))
+        ;; Return keyword
+        ((eql (char symbol-string 0) #\:)
+         (return-from find-best-symbol
+           (values nil
+                   (intern-keyword (subseq symbol-string 1))
+                   rest-string)))
+        ((and (not (null symbol)) (symbolp symbol))
+         (if (member symbol ignore-symbols)
+             (return-from find-best-symbol
+               (values nil symbol rest-string))
+             (return-from find-best-symbol
+               (values symbol nil rest-string)))))))
+  (values nil nil (car symbols)))
 
 (defun scope-symbols-in-text (text &optional ignore-symbols)
   (flet ((whitespace-p (char)
@@ -154,8 +187,8 @@ possible symbol names."
               ;; if probably an example and peek eol, then switch back.
               ((and probably-example
                     (peek-char nil stream nil)
-                    (eql (peek-char nil stream nil) #\Newline)
-                    (setf probably-example nil))
+                    (eql (peek-char nil stream nil) #\Newline))
+               (setf probably-example nil)
                (write-char char out))
               ;; new line, then indentation, probably an example.
               ((and (eql char #\Newline)
@@ -186,10 +219,12 @@ possible symbol names."
               ((and possible-symbol (not (upper-case-p char)))
                (push (coerce (reverse possible-symbol) 'string)
                      possible-symbols)
-               (multiple-value-bind (symbol rest)
+               (multiple-value-bind (symbol literal rest)
                    (find-best-symbol possible-symbols ignore-symbols)
                  (when symbol
                    (write-string (encode-xref symbol) out))
+                 (when literal
+                   (write-string (encode-literal literal) out))
                  (write-string rest out))
                (unread-char char stream)
                (setf possible-symbol nil)
@@ -205,8 +240,17 @@ possible symbol names."
               ;; else just write the character out
               (t
                (write-char char out))))
+          ;; Write out remaining symbols
           (when possible-symbol
-            (write-string (coerce (reverse possible-symbol) 'string) out)))))))
+            (push (coerce (reverse possible-symbol) 'string)
+                  possible-symbols)
+            (multiple-value-bind (symbol literal rest)
+                   (find-best-symbol possible-symbols ignore-symbols)
+                 (when symbol
+                   (write-string (encode-xref symbol) out))
+                 (when literal
+                   (write-string (encode-literal literal) out))
+                 (write-string rest out))))))))
 
 (defun arglist (symbol)
   #+sbcl
@@ -246,23 +290,31 @@ possible symbol names."
   (:documentation
    "Encode documentation for a symbol as a JSON object member."))
 
+(defgeneric encode-function-documentation (symbol type)
+  (:documentation
+   "Encode documentation for a symbol as a JSON object member."))
+
 (defmethod encode-function-documentation (symbol (type (eql 'function)))
   (encode-function-documentation*
    symbol type (or (documentation symbol type) "")))
 
 (defmethod encode-function-documentation (symbol (type (eql 'macro)))
   (encode-function-documentation*
-   symbol type (or (documentation symbol type) "")))
+   symbol type (or (documentation symbol 'function) "")))
 
 (defmethod encode-function-documentation (symbol (type (eql 'generic-function)))
   (encode-function-documentation*
-   symbol type (or (documentation symbol type) ""))
+   symbol type (or (documentation symbol 'function) ""))
   (as-object-member ('methods)
     (encode-methods (symbol-function symbol))))
 
-(defmethod encode-function-documentation (symbol (type (eql 'cl:STANDARD-GENERIC-FUNCTION)))
+(defmethod encode-function-documentation (symbol (type (eql 'cl:standard-generic-function)))
   (encode-function-documentation symbol 'generic-function))
 
+(defgeneric encode-value-documentation (symbol type)
+  (:documentation
+   "Encode documentation for a symbol as a JSON object member."))
+
 (defmethod encode-value-documentation (symbol (type (eql 'variable)))
   (encode-variable-documentation* symbol type))
 
@@ -322,7 +374,8 @@ possible symbol names."
 
 (defun main ()
   (multiple-value-bind (unused-args args invalid-args)
-      (getopt:getopt (argv) '(("package" :required)("path" :required)))
+      (getopt:getopt (argv) '(("package" :required)
+                              ("path" :required)))
     (cond
       (invalid-args
        (print-message "Invalid arguments")
